// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"crypto/md5" // #nosec
	"encoding/hex"
	"io"
	"os"
	"time"
)

const hashExpiration = 24 * 3600 // 1 day in seconds
const FastHashSize = 8192        // 8 KB

type fileStatInfo struct {
	path      string
	info      os.FileInfo
	opened    *os.File
	lastError error
}

func newFileStatInfo(path string) *fileStatInfo {
	s := &fileStatInfo{path: path}

	info, err := os.Stat(path)
	if err != nil {
		s.lastError = err
		return s
	}

	s.info = info

	return s
}

func (f *fileStatInfo) close() {
	if f.opened != nil {
		_ = f.opened.Close()
	}
}

func (f *fileStatInfo) open() {
	if f.opened != nil {
		return
	}

	f.opened, f.lastError = os.Open(f.path)
}

func (f *fileStatInfo) exists() bool {
	return f.info != nil
}

var (
	// watcherImpl is the singleton instance of the Watcher.
	watcherImpl *Watcher
)

func init() {
	watcherImpl = NewWatcher()
}

// buildHash generates both a complete and a "fast" hash.
// the fast hash is for hashing big files.
func buildHash(info *fileStatInfo) FileHash {
	h := FileHash{Path: info.path}

	hash, fastHash := hashFile(info)
	if info.lastError != nil {
		return h
	}

	h.Hash = hash
	h.FastHash = fastHash
	h.expiresAt = time.Now().Add(hashExpiration * time.Second)
	h.lastModTime = info.info.ModTime()
	return h
}

// hashFile creates a full hash and a "fast" hash.
func hashFile(info *fileStatInfo) (*string, *string) {
	// Full hash
	fullHash := fullHashFile(info)
	if info.lastError != nil {
		return nil, nil
	}

	// Fast hash
	fastHash := fastHashFile(info)
	if info.lastError != nil {
		return nil, nil
	}

	return fullHash, fastHash
}

// fullHashFile creates a complete hash.
func fullHashFile(info *fileStatInfo) *string {

	info.open()
	if info.lastError != nil {
		return nil
	}

	file := info.opened

	// #nosec
	hash := md5.New()
	if _, err := io.Copy(hash, file); err != nil {
		info.lastError = err
		return nil
	}

	if _, err := file.Seek(0, 0); err != nil {
		info.lastError = err
		return nil
	}

	hashString := hex.EncodeToString(hash.Sum(nil)[:16])
	return &hashString
}

// fastHashFile creates a "fast" hash.
func fastHashFile(info *fileStatInfo) *string {
	info.open()
	if info.lastError != nil {
		return nil
	}

	file := info.opened

	fileSize := info.info.Size()
	if fileSize < FastHashSize {
		return fullHashFile(info)
	}

	offset := fileSize/2 - FastHashSize/2

	// #nosec
	fastHash := md5.New()
	buffer := make([]byte, FastHashSize)

	readPositions := []int64{0, offset, fileSize - FastHashSize}
	for _, pos := range readPositions {
		if _, err := file.ReadAt(buffer, pos); err != nil {
			return nil
		}
		fastHash.Write(buffer)
	}

	hashString := hex.EncodeToString(fastHash.Sum(nil)[:16])
	return &hashString
}
