// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"os"
	"time"
)

// FileHash contains the md5 hash of a file and the last modification time.
type FileHash struct {
	Path        string
	Hash        *string
	FastHash    *string
	expiresAt   time.Time
	lastModTime time.Time
}

// RemoveFileHash removes a file from the fileHashCache.
// if file not in cache return error that file not in cache
// This function uses the global fileHashCache.
func RemoveFileHash(path string) error {
	return watcherImpl.Remove(path)
}

// Remove removes a file from the fileHashCache.
// If the watcher is not active, it returns an error (WatcherNotActiveError).
func (w *Watcher) Remove(path string) error {

	if !w.IsActive() {
		return WatcherNotActiveError{}
	}

	// check if file in fileHashCache, if not return error that file not in cache
	if _, ok := w.cache.Load(path); !ok {
		return os.ErrNotExist
	}

	w.cache.Delete(path)
	return nil
}

// AddFileHash adds a file to the fileHashCache.
// This function uses the global fileHashCache.
func AddFileHash(path string) error {
	return watcherImpl.Add(path)
}

// Add adds a file to the watcher cache.
func (w *Watcher) Add(path string) error {

	if !w.IsActive() {
		return WatcherNotActiveError{}
	}

	info := newFileStatInfo(path)
	defer info.close()

	if !info.exists() {
		return os.ErrNotExist
	}

	h := buildHash(info)
	if info.lastError != nil {
		return info.lastError
	}

	w.cache.Store(path, h)
	return nil
}

// HasFileChanged checks if a file has changed since the last time it was checked.
// It returns true if the file has changed, false if it hasn't and an error if
// something went wrong.
// This function uses the global fileHashCache.
func HasFileChanged(path string) (bool, error) {
	return watcherImpl.Has(path)
}

// Has checks if a file has changed since the last time it was checked.
// It returns true if the file has changed, false if it hasn't and an error if
// something went wrong.
// If the watcher is not active, it returns false and an error (WatcherNotActiveError).
func (w *Watcher) Has(path string) (bool, error) {

	if !w.IsActive() {
		return false, WatcherNotActiveError{}
	}

	value, ok := w.cache.Load(path)
	if !ok {
		return true, nil
	}

	cachedHash, ok := value.(FileHash)
	if !ok {
		return true, nil
	}

	info := newFileStatInfo(path)
	defer info.close()

	if !info.exists() {
		return true, nil
	}

	modTime := info.info.ModTime()
	if modTime.After(cachedHash.lastModTime) {
		return true, nil
	}

	currentFastHash := fastHashFile(info)
	if info.lastError != nil {
		return false, info.lastError
	}

	if currentFastHash != nil && cachedHash.FastHash != nil && *currentFastHash != *cachedHash.FastHash {
		return true, nil
	}

	// Calculate full hash only if fast hash differs
	currentFullHash := fullHashFile(info)
	if info.lastError != nil {
		return false, info.lastError
	}

	if *currentFullHash != *cachedHash.Hash {
		return true, nil
	}

	return false, nil
}
