// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"os"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func TestAddFileHash(t *testing.T) {
	// Create a temporary test file
	file, err := os.CreateTemp("", "testfile")
	if err != nil {
		t.Fatal("Failed to create test file:", err)
	}
	defer func() {
		_ = os.Remove(file.Name())
	}()

	_ = file.Close()

	err = AddFileHash(file.Name())
	assert.Nil(t, err)

	_, ok := watcherImpl.cache.Load(file.Name())
	assert.True(t, ok)
}

func TestRemoveFileHash(t *testing.T) {
	watcherImpl.cache.Store("testfile", FileHash{Path: "testfile"})
	err := RemoveFileHash("testfile")
	assert.Nil(t, err)

	_, ok := watcherImpl.cache.Load("testfile")
	assert.False(t, ok)
}

func TestRemoveFileHash_NotExist(t *testing.T) {
	err := RemoveFileHash("nonexistent")
	assert.Equal(t, os.ErrNotExist, err)
}

func TestHasFileChanged(t *testing.T) {
	// Create a temporary test file
	file, err := os.CreateTemp("", "testfile")
	if err != nil {
		t.Fatal("Failed to create test file:", err)
	}
	defer func() {
		_ = os.Remove(file.Name())
	}()
	_ = file.Close()

	// Add it to the cache with a specific lastModTime
	watcherImpl.cache.Store(file.Name(), FileHash{Path: file.Name(), lastModTime: time.Now().Add(-1 * time.Hour)})

	// Test Has
	changed, err := HasFileChanged(file.Name())
	assert.Nil(t, err)
	assert.True(t, changed)
}
func TestHasFileChanged_NotExist(t *testing.T) {
	changed, err := HasFileChanged("nonexistent")
	assert.Nil(t, err)
	assert.True(t, changed)
}

func TestHasFileChanged_InvalidCacheType(t *testing.T) {
	watcherImpl.cache.Store("testfile", "invalid type")
	changed, err := HasFileChanged("testfile")
	assert.Nil(t, err)
	assert.True(t, changed)
}

func TestAddFileHashWithInvalidPath(t *testing.T) {
	err := AddFileHash("/invalid/path")
	if err == nil {
		t.Errorf("Expected an error for invalid path")
	}
}

func TestAddAndRemoveFileHash2(t *testing.T) {
	file, err := os.CreateTemp("", "testfile")
	if err != nil {
		t.Fatal("Failed to create test file:", err)
	}
	defer func() {
		_ = os.Remove(file.Name())
	}()
	_ = file.Close()

	err = AddFileHash(file.Name())
	if err != nil {
		t.Errorf("Error adding valid path: %s", err)
	}

	err = RemoveFileHash(file.Name())
	if err != nil {
		t.Errorf("Error removing valid path: %s", err)
	}
}

func TestAddFileHashRaceCondition(t *testing.T) {
	for i := 0; i < 100; i++ {
		go func() {
			_ = AddFileHash("/some/path")
		}()
	}
}

func TestRemoveFileHashRaceCondition(t *testing.T) {
	for i := 0; i < 100; i++ {
		go func() {
			_ = RemoveFileHash("/some/path")
		}()
	}
}

func TestHasFileChangedRaceCondition(t *testing.T) {
	for i := 0; i < 100; i++ {
		go func() {
			_, _ = HasFileChanged("/some/path")
		}()
	}
}

func BenchmarkAddFileHash(b *testing.B) {
	b.ReportAllocs()

	for i := 0; i < b.N; i++ {
		file, err := os.CreateTemp("", "testfile")
		if err != nil {
			b.Fatal("Failed to create test file:", err)
		}

		_ = AddFileHash(file.Name())

		if err := RemoveFileHash(file.Name()); err != nil {
			b.Fatal("Failed to remove file hash:", err)
		}

		if err := os.Remove(file.Name()); err != nil {
			b.Fatal("Failed to remove test file:", err)
		}
	}
}

func BenchmarkHasFileChanged(b *testing.B) {
	b.ReportAllocs()

	file, err := os.CreateTemp("", "testfile")
	if err != nil {
		b.Fatal("Failed to create test file:", err)
	}
	_ = AddFileHash(file.Name())

	b.ResetTimer()

	for i := 0; i < b.N; i++ {

		// change the file every 100 iterations
		if i%100 == 0 {
			_, _ = file.WriteString("test")
		}

		_, _ = HasFileChanged(file.Name())
	}

}
