// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"os"
	"path/filepath"
	"testing"
	"time"
)

// getOrCreateFileHash returns the md5 hash of a file. If the hash is not in the cache,
// it will be calculated. If the file does not exist, an error is returned.
// This is a helper function for the tests.
func getOrCreateFileHash(info *fileStatInfo) (*string, *string) {
	if !info.exists() {
		return nil, nil
	}

	p := info.path
	if value, ok := watcherImpl.cache.Load(p); ok {
		hash := value.(FileHash)
		if time.Now().Before(hash.expiresAt) {
			return hash.Hash, hash.FastHash
		}
	}

	h := buildHash(info)
	if info.lastError != nil {
		return nil, nil
	}

	watcherImpl.cache.Store(p, h)
	return h.Hash, h.FastHash
}

func TestFileHashing(t *testing.T) {
	// Create a temporary file
	tempFile, err := os.CreateTemp("", "tempFile")
	if err != nil {
		t.Fatalf("Cannot create temporary file: %s", err)
	}
	defer func() {
		_ = os.Remove(tempFile.Name())
	}()
	// Write data to the temporary file
	if _, err := tempFile.Write([]byte("Hello, world!")); err != nil {
		t.Fatalf("Cannot write to temporary file: %s", err)
	}
	_ = tempFile.Close()

	// First hashing
	info := newFileStatInfo(tempFile.Name())
	hash1, fastHash1 := getOrCreateFileHash(info)
	if info.lastError != nil {
		t.Fatalf("Error hashing file: %s", err)
	}

	// Open again and change data
	tempFile, err = os.OpenFile(tempFile.Name(), os.O_WRONLY|os.O_APPEND, 0666)
	if err != nil {
		t.Fatalf("Cannot open temporary file: %s", err)
	}
	if _, err := tempFile.Write([]byte(" More data.")); err != nil {
		t.Fatalf("Cannot write to temporary file: %s", err)
	}
	_ = tempFile.Close()

	// Wait to make sure that the timestamps are different
	time.Sleep(2 * time.Second)

	// Second hashing
	info = newFileStatInfo(tempFile.Name())
	hash2, fastHash2 := getOrCreateFileHash(info)
	if info.lastError != nil {
		t.Fatalf("Error hashing file: %s", err)
	}

	// Check the hashes
	if hash1 != hash2 {
		t.Errorf("Expected full hashes not to differ: %s == %s", *hash1, *hash2)
	}

	if fastHash1 != fastHash2 {
		t.Errorf("Expected fast hashes not to differ: %s == %s", *fastHash1, *fastHash2)
	}

	// Check if the file has been changed
	hasChanged, err := HasFileChanged(tempFile.Name())
	if err != nil {
		t.Fatalf("Error checking file change: %s", err)
	}

	if !hasChanged {
		t.Errorf("Expected Has to return true, got false")
	}
}
func TestAddAndRemoveFileHash(t *testing.T) {
	// Test cases
	testCases := []struct {
		fileName    string
		shouldExist bool
		sleepTime   time.Duration
		expectedErr error
	}{
		{"testfile1.txt", true, 0, nil},
		{"testfile2.txt", true, 0, nil},
		{"testfile3.txt", false, 0, os.ErrNotExist},
		{"testfile1.txt", true, 2 * time.Second, nil}, // Test expiration
	}

	// Create a temporary directory for test files
	tempDir, err := os.MkdirTemp("", "testdir")
	if err != nil {
		t.Fatalf("Failed to create temp directory: %v", err)
	}
	defer os.RemoveAll(tempDir) // Clean up the temp directory at the end

	// Create test files in the temp directory
	for _, tc := range testCases {
		if tc.shouldExist {
			filePath := filepath.Join(tempDir, tc.fileName)
			file, err := os.Create(filePath)
			if err != nil {
				t.Fatal(err)
			}
			_ = file.Close()
		}
	}

	for _, tc := range testCases {
		filePath := filepath.Join(tempDir, tc.fileName)

		//Add hash
		err := AddFileHash(filePath)
		if err != tc.expectedErr {
			t.Errorf("Add(%q) error: got %v, want %v", filePath, err, tc.expectedErr)
		}

		if tc.sleepTime > 0 {
			time.Sleep(tc.sleepTime)
		}

		// Remove hash
		err = RemoveFileHash(filePath)
		if err != tc.expectedErr {
			t.Errorf("Remove(%q) error: got %v, want %v", filePath, err, tc.expectedErr)
		}

		_, ok := watcherImpl.cache.Load(filePath)
		if ok {
			t.Errorf("Remove(%q) failed: value still exists in cache", filePath)
		}
	}
}
