// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"fmt"
	"github.com/fsnotify/fsnotify"
	"os"
	"path/filepath"
	"sync"
	"time"
)

type Watch struct {
	Path     string
	OnCreate EventCallback
	OnChange EventCallback
	OnDelete EventCallback
	OnRename EventCallback
	Tags     []string
}

type EventErrorCallback func(err error)
type EventCallback func(path string)

type lighthouse struct {
	watchers map[string]*Watch
	fsnotify *fsnotify.Watcher
	mutex    sync.Mutex
	active   bool
	running  bool
	onError  EventErrorCallback
	debounce time.Duration
	quit     chan struct{}
}

// SetOnError Methode, um onError zu setzen
func (l *lighthouse) SetOnError(callback EventErrorCallback) {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	l.onError = callback
}

// SetDebounce Methode, um debounce zu setzen
func (l *lighthouse) SetDebounce(duration time.Duration) {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	l.debounce = duration
}

type Lighthouse interface {
	Add(watch *Watch) error
	Remove(path string) error
	IsActiveWatched(path string) bool
	IsWatched(path string) bool
	IsRunning() bool

	IsActive() bool
	StartWatching() error

	StopWatching() error

	SetOnError(callback EventErrorCallback)
	SetDebounce(duration time.Duration)

	WatchList() []string
	Sync() error

	IsInSync() bool

	WatchListByTags(tags []string) []string

	RemoveByTags(tags []string) error
}

// NewLighthouse creates a new lighthouse instance.
func NewLighthouse() Lighthouse {
	l := &lighthouse{}
	l.debounce = 500 * time.Millisecond
	l.checkAndInit()
	return l

}

func (l *lighthouse) checkAndInit() {

	var err error

	if l.active {
		return
	}

	if l.watchers == nil {
		l.watchers = make(map[string]*Watch)
	}

	if l.fsnotify == nil {
		l.fsnotify, err = fsnotify.NewWatcher()
		if err != nil {
			return
		}
	}

	// test if fsnotify is running
	err = l.fsnotify.Add("/undefined-path-123456789")
	if err != nil {
		if err.Error() == "inotify instance already closed" {
			l.fsnotify, err = fsnotify.NewWatcher()
			if err != nil {
				return
			}
		}
	}

	l.active = true

}

// WatchList returns a list of all watched paths.
func (l *lighthouse) WatchList() []string {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	var list []string
	for k := range l.watchers {
		list = append(list, k)
	}
	return list

}

// WatchListByTags returns a list of all watched paths with the given tags.
func (l *lighthouse) WatchListByTags(tags []string) []string {

	if len(tags) == 0 {
		return l.WatchList()
	}

	l.mutex.Lock()
	defer l.mutex.Unlock()
	var list []string
	for k := range l.watchers {

		if len(l.watchers[k].Tags) == 0 {
			continue
		}

		for _, tag := range tags {
			if tag == k {
				list = append(list, k)
				break
			}
		}
	}
	return list

}

// IsInSync returns true if all paths are watched by fsnotify.
func (l *lighthouse) IsInSync() bool {
	l.mutex.Lock()
	defer l.mutex.Unlock()

	wl := l.fsnotify.WatchList()

	if len(wl) != len(l.watchers) {
		return false
	}

	for _, w := range wl {
		if _, ok := l.watchers[w]; !ok {
			return false
		}
	}

	return true

}

// Sync Lighthouse and fsnotify. Remove all watchers from fsnotify and add them again.
func (l *lighthouse) Sync() error {
	l.mutex.Lock()
	defer l.mutex.Unlock()

	// remove all watchers from fsnotify
	wl := l.fsnotify.WatchList()
	for _, w := range wl {
		_ = l.fsnotify.Remove(w)
	}

	// add all watchers to fsnotify
	var errReturn error

	for k := range l.watchers {
		err := l.fsnotify.Add(k)
		if err != nil {
			if errReturn == nil {
				errReturn = err
			} else {
				errReturn = fmt.Errorf("%w, %v", errReturn, err)
			}
		}
	}

	return errReturn
}

// IsRunning returns true if the watcher is active, false otherwise.
func (l *lighthouse) IsRunning() bool {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	return l.running
}

// IsActive returns true if the watcher is active, false otherwise.
func (l *lighthouse) IsActive() bool {
	l.mutex.Lock()
	defer l.mutex.Unlock()
	return l.active
}

// normalizePath returns the normalized path. If the path is a directory, a
// trailing slash is added.
// If the path does not exist, an error is returned.
// If the path is a directory, the second return value is true.
func normalizePath(path string) (string, bool, error) {

	isDir := false

	fi, err := os.Stat(path)
	if err != nil {
		return "", isDir, err
	}

	// normalize is path is directory
	if fi.IsDir() {
		path = filepath.Clean(path)
		isDir = true
	}

	return path, isDir, nil

}

// Add adds a path to the watcher. If the path is already being watched or the
// watcher is not active, an error is returned.
func (l *lighthouse) Add(watch *Watch) error {

	l.mutex.Lock()
	defer l.mutex.Unlock()

	var err error

	if !l.active {
		return LighthouseNotActiveError{}
	}

	if _, ok := l.watchers[watch.Path]; ok {
		return AlreadyWatchedPathError{AlreadyWatched: watch.Path}
	}

	watch.Path, _, err = normalizePath(watch.Path)
	if err != nil {
		return err
	}

	l.watchers[watch.Path] = watch
	err = l.fsnotify.Add(watch.Path)
	if err != nil {
		return err
	}

	return nil
}

func (l *lighthouse) RemoveByTags(tags []string) error {

	if len(tags) == 0 {
		return nil
	}

	l.mutex.Lock()
	defer l.mutex.Unlock()

	var errReturn error

	for _, tag := range tags {

		for k := range l.watchers {

			if len(l.watchers[k].Tags) == 0 {
				continue
			}

			for _, t := range l.watchers[k].Tags {
				if t == tag {
					err := l.fsnotify.Remove(k)
					if err != nil {
						if errReturn == nil {
							errReturn = err
						} else {
							errReturn = fmt.Errorf("%w, %v", errReturn, err)
						}
					}
					delete(l.watchers, k)
					break
				}
			}
		}
	}

	return errReturn

}

// Remove removes a path from the watcher. If the path is not being watched or
// the watcher is not active, an error is returned.
func (l *lighthouse) Remove(path string) error {
	l.mutex.Lock()
	defer l.mutex.Unlock()

	if !l.active {
		return LighthouseNotActiveError{}
	}

	path, _, _ = normalizePath(path)

	if _, ok := l.watchers[path]; !ok {
		return UnwatchedPathError{UnwatchedPath: path}
	}

	err := l.fsnotify.Remove(path)
	delete(l.watchers, path)

	if err != nil {
		return err
	}

	return nil
}

// IsWatched returns true if the path is being watched, false otherwise.
// This only checks the lighthouse, not the fsnotify watcher.
// If you want to check if the path is being watched by fsnotify, use IsActiveWatched.
func (l *lighthouse) IsWatched(path string) bool {
	l.mutex.Lock()
	defer l.mutex.Unlock()

	if !l.active {
		return false
	}

	if l.watchers == nil {
		return false
	}

	if l.fsnotify == nil {
		return false
	}

	path, _, _ = normalizePath(path)

	if _, ok := l.watchers[path]; !ok {
		return false
	}

	return true
}

// Get returns the watch for the given path. If the path is not being watched
// or the watcher is not active, an error is returned.
func (l *lighthouse) Get(path string) (*Watch, error) {
	l.mutex.Lock()
	defer l.mutex.Unlock()

	if !l.active {
		return nil, LighthouseNotActiveError{}
	}

	var watch *Watch
	var ok bool
	var isDir bool

	path, isDir, _ = normalizePath(path)

	if watch, ok = l.watchers[path]; !ok {
		if isDir {
			return nil, UnwatchedPathError{UnwatchedPath: path}
		}

		dir := filepath.Dir(path)
		if watch, ok = l.watchers[dir]; !ok {
			return nil, UnwatchedPathError{UnwatchedPath: path}
		}

	}

	return watch, nil
}

// IsActiveWatched returns true if the path is being watched, false otherwise.
// If the watcher is not active, false is returned.
// If the path is in the watcher but not in the fsnotify watcher, false is returned.
func (l *lighthouse) IsActiveWatched(path string) bool {

	if l.IsWatched(path) == false {
		return false
	}

	l.mutex.Lock()
	defer l.mutex.Unlock()

	path, _, _ = normalizePath(path)

	// check fsnotify
	for _, fsPath := range l.fsnotify.WatchList() {
		if fsPath == path {
			return true
		}

	}

	return false

}
