// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"github.com/fsnotify/fsnotify"
	"os"
	"sync"
	"time"
)

// StopWatching stops watching the given file for changes
// If the watcher is not running, an error WatcherNotRunningError is returned.
func (l *lighthouse) StopWatching() error {

	if !l.IsRunning() {
		return WatcherNotRunningError{}
	}

	l.mutex.Lock()
	defer l.mutex.Unlock()

	close(l.quit)
	l.running = false
	return nil
}

// StartWatching starts watching the given file for changes
func (l *lighthouse) StartWatching() error {
	if l.IsRunning() {
		return WatcherAlreadyRunningError{}
	}

	l.mutex.Lock()
	defer l.mutex.Unlock()
	l.quit = make(chan struct{})
	l.running = true
	l.checkAndInit()

	go func() {
		eventChannel := make(chan fsnotify.Event, 100)
		errorChannel := make(chan error, 100)
		debounceTimers := make(map[string]map[fsnotify.Op]*time.Timer)

		var debounceMutex sync.Mutex

		go func() {
			for {

				l.mutex.Lock()
				quit := l.quit
				l.mutex.Unlock()

				select {
				case <-quit:
					return
				case event := <-l.fsnotify.Events:
					eventChannel <- event
				case err := <-l.fsnotify.Errors:
					errorChannel <- err
				}
			}
		}()

		for {

			l.mutex.Lock()
			quit := l.quit
			l.mutex.Unlock()

			select {
			case <-quit:
				return
			case event := <-eventChannel:

				debounceMutex.Lock()
				if _, ok := debounceTimers[event.Name]; !ok {
					debounceTimers[event.Name] = make(map[fsnotify.Op]*time.Timer)
				}

				if timer, ok := debounceTimers[event.Name][event.Op]; ok {
					timer.Stop()
				}

				debounceTimers[event.Name][event.Op] = time.AfterFunc(l.debounce, func() {
					debounceMutex.Lock()
					delete(debounceTimers[event.Name], event.Op)
					if len(debounceTimers[event.Name]) == 0 {
						delete(debounceTimers, event.Name)
					}
					debounceMutex.Unlock()

					var watch *Watch
					var err error

					if watch, err = l.Get(event.Name); err != nil || watch == nil {
						return
					}

					switch event.Op {
					case fsnotify.Write:
						if watch.OnChange != nil {
							if watch.OnChange != nil {
								go watch.OnChange(event.Name)
							}
						}
					case fsnotify.Create:
						if watch.OnCreate != nil {
							if watch.OnCreate != nil {
								go watch.OnCreate(event.Name)
							}
						}

					case fsnotify.Rename:
						if watch.OnRename != nil {
							if watch.OnRename != nil {
								go watch.OnRename(event.Name)
							}
						}

						go func(filename string) {
							ticker := time.NewTicker(1 * time.Second)
							for {
								select {
								case <-ticker.C:
									if l.IsWatched(filename) {
										if _, err := os.Stat(filename); err == nil {
											_ = l.fsnotify.Add(filename)
											ticker.Stop()
											return
										}
									} else {
										ticker.Stop()
										return
									}
								}
							}
						}(event.Name)

					case fsnotify.Remove:
						if watch.OnDelete != nil {
							if watch.OnDelete != nil {
								go watch.OnDelete(event.Name)
							}
						}

						go func(filename string) {
							ticker := time.NewTicker(1 * time.Second)

							for {
								select {
								case <-ticker.C:

									if l.IsWatched(filename) {
										if _, err := os.Stat(filename); err == nil {
											_ = l.fsnotify.Add(filename)
											ticker.Stop()
											return
										}
									} else {
										ticker.Stop()
										return
									}

								}
							}
						}(event.Name)

					}

				})
				debounceMutex.Unlock()

			case err := <-errorChannel:
				if l.onError != nil {
					go l.onError(err)
				}
			}
		}
	}()

	return nil
}
