// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package watch

import (
	"github.com/stretchr/testify/assert"
	"os"
	"path/filepath"
	"testing"
	"time"
)

func TestWriteEvent(t *testing.T) {
	tmpDir := t.TempDir()
	t.Cleanup(func() {
		_ = os.RemoveAll(tmpDir) // Cleanup
	})
	filePath := filepath.Join(tmpDir, "test.txt")

	l := NewLighthouse()

	ch := make(chan bool)
	timeout := make(chan bool)

	t.Cleanup(func() {
		close(ch)
		close(timeout)
	})

	// Create file and add watcher
	if err := os.WriteFile(filePath, []byte("test"), 0644); err != nil {
		t.Fatalf("Failed to write file: %v", err)
	}

	w1 := &Watch{
		Path: tmpDir,
		OnChange: func(name string) {
			ch <- true
		},
	}

	if err := l.Add(w1); err != nil {
		t.Fatalf("Failed to add watch: %v", err)
	}

	l.StartWatching()

	timer := time.NewTimer(5 * time.Second)
	go func() {
		<-timer.C
		timeout <- true
	}()

	// Change file
	if err := os.WriteFile(filePath, []byte("test2"), 0644); err != nil {
		t.Fatalf("Failed to write file: %v", err)
	}

	select {
	case <-ch:
		timer.Stop()
	case <-timeout:
		t.Fail()
	}
}

func TestRecreateEvent(t *testing.T) {
	tmpDir := t.TempDir()
	t.Cleanup(func() {
		_ = os.RemoveAll(tmpDir)
	})
	filePath := filepath.Join(tmpDir, "recreate.txt")

	callCounterCreate := 0
	callCounterChange := 0

	l := NewLighthouse()
	ch := make(chan bool)
	timeout := make(chan bool)

	t.Cleanup(func() {
		close(ch)
		close(timeout)
	})

	w1 := &Watch{
		Path: tmpDir,
		OnChange: func(name string) {
			callCounterChange++
			ch <- true
		},
		OnCreate: func(name string) {
			callCounterCreate++
			ch <- true
		},
	}

	if err := l.Add(w1); err != nil {
		t.Fatalf("Failed to add watch: %v", err)
	}

	_ = l.StartWatching()

	timer := time.NewTimer(20 * time.Second)
	go func() {
		<-timer.C
		timeout <- true
	}()

	// Create file
	if err := os.WriteFile(filePath, []byte("test1"), 0644); err != nil {
		t.Fatalf("Failed to write file: %v", err)
	}

	select {
	case <-ch:
		// First create event detected
	case <-timeout:
		t.Log("Timeout 1")
		t.Fail()
		return
	}

	// Delete the file after 2 seconds
	time.Sleep(2 * time.Second)
	if err := os.Remove(filePath); err != nil {
		t.Fatalf("Failed to remove file: %v", err)
	}
	time.Sleep(2 * time.Second)

	// Recreate the file after another 2 seconds
	if err := os.WriteFile(filePath, []byte("test2"), 0644); err != nil {
		t.Fatalf("Failed to write file: %v", err)
	}

	select {
	case <-ch:
		timer.Stop()

		assert.Equal(t, 1, callCounterCreate)
		assert.Equal(t, 1, callCounterChange)

	case <-timeout:
		t.Log("Timeout 2")
		t.Fail()
	}
}

func TestStopWatching(t *testing.T) {
	l := NewLighthouse() // Erstellen Sie eine neue Instanz der `lighthouse`-Struktur.

	// Start Watching
	_ = l.StartWatching()

	// Erlauben Sie dem System etwas Zeit, um den Watch-Prozess zu starten.
	time.Sleep(time.Second * 1)
	if !l.IsRunning() {
		t.Fatal("l.active is false, but should be true after StartWatching")
	}

	// Stop Watching
	err := l.StopWatching()
	if err != nil {
		t.Fatalf("StopWatching failed: %v", err)
	}

	// Überprüfen Sie, ob die Goroutine und der Watcher gestoppt wurden.
	if l.IsRunning() {
		t.Fatal("l.active is true, but should be false after StopWatching")
	}

	_ = l.StartWatching()
	assert.True(t, l.IsRunning())

}
